package simulation;

import metadata.BuildingMetadata;
import metadata.FactoryInput;
import metadata.FactoryMetadata;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

public class Factory extends Building implements WarehouseObserver {
    private final FactoryMetadata metadata;
    private final Map<ComponentType, Integer> inputsCount = new HashMap<>();

    private float productionInterval;
    private boolean isProductionStarted;
    private boolean haltProduction;
    private int productionTicks = 0;

    public Factory(int id, String type, int x, int y, FactoryMetadata metadata) {
        super(id, type, x, y);
        this.metadata = metadata;
        this.productionInterval = metadata.getProductionInterval();

        // Les usines sans entrées doivent toujours produire
        isProductionStarted = metadata.getInputs().size() == 0;
        if (isProductionStarted) {
            productionTicks = (int) productionInterval;
        }
    }

    @Override
    public Optional<Component> produce() {
        if (!isProductionStarted ||
                (productionTicks == 0 && haltProduction)) { // La production en cours ne doit pas s'arrêter, mais l'usine ne doit pas produire un nouveau composant
            return Optional.empty();
        }

        productionTicks++;
        if (productionTicks < productionInterval) {
            // La production n'est pas finie
            updateState();
            return Optional.empty();
        }

        // La production est finie
        // Réinitialise la progression de la production et défini l'état de l'usine à vide.
        // Retourne le composant produit
        if (metadata.getInputs().size() > 0) {
            isProductionStarted = false;
        }

        productionTicks = 0;
        state = BuildingState.EMPTY;
        Component newComponent = new Component(metadata.getOutput().getType());
        return Optional.of(newComponent);
    }

    @Override
    public void processInput(Component input) {
        int inputCount = 0;

        // S'assure qu'une entrée pour ce type de composant existe.
        if (inputsCount.containsKey(input.getType())) {
            inputCount = inputsCount.get(input.getType());
        }

        // Augmente la quantité de ce type de composant en stock.
        inputsCount.put(input.getType(), inputCount + 1);

        if (hasEnoughComponents()) {
            startProduction();
        }
    }

    @Override
    public BuildingMetadata getMetadata() {
        return metadata;
    }

    @Override
    public void updateWarehouseState(BuildingState state) {
        updateProductionSpeed(state);
    }

    private void updateProductionSpeed(BuildingState state) {
        int originalProductionInterval = metadata.getProductionInterval();

        // Augmente le temps de production selon l'état de l'entrepôt
        switch (state) {
            case EMPTY -> {
                haltProduction = false;
                productionInterval = originalProductionInterval;
            }
            case ONE_THIRD -> {
                haltProduction = false;
                productionInterval = originalProductionInterval * 1.33333f;
            }
            case TWO_THIRD -> {
                haltProduction = false;
                productionInterval = originalProductionInterval * 2.66666f;
            }
            default -> haltProduction = true;
        }
    }

    private void updateState() {
        float productionRatio = productionTicks / productionInterval;
        if (productionRatio >= (2 / 3f)) {
            state = BuildingState.FULL;
        } else if (productionRatio >= (1 / 3f)) {
            state = BuildingState.TWO_THIRD;
        } else {
            state = BuildingState.ONE_THIRD;
        }
    }

    private boolean hasEnoughComponents() {
        for (FactoryInput input : metadata.getInputs()) {
            if (!inputsCount.containsKey(input.getType()) ||
                    inputsCount.get(input.getType()) < input.getQuantity()) {
                return false;
            }
        }

        return true;
    }

    private void startProduction() {
        // Retire les composants nécessaires des entrées
        for (FactoryInput input : metadata.getInputs()) {
            int inputCount = inputsCount.get(input.getType());
            inputsCount.put(input.getType(), inputCount - input.getQuantity());
        }

        isProductionStarted = true;
        productionTicks = 0;
    }

    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder("[");

        if (!inputsCount.isEmpty()) {
            for (ComponentType type : inputsCount.keySet()) {
                builder.append(type.toString()).append(": ").append(inputsCount.get(type));
            }
        }

        if (isProductionStarted) {
            builder.append(" (interval: ").append(productionTicks).append(')');
        }

        builder.append(']');
        return builder.toString();
    }
}
