package simulation;

import metadata.BuildingMetadata;
import metadata.WarehouseMetadata;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Optional;

public class Warehouse extends Building implements WarehouseSubject {
    private final WarehouseMetadata metadata;
    private final int capacity;
    private final Collection<WarehouseObserver> observers = new ArrayList<>();

    private SellStrategy sellStrategy = new TimedSellStrategy();
    private int planeCount = 0;

    public Warehouse(int id, String type, int x, int y, WarehouseMetadata metadata) {
        super(id, type, x, y);

        this.metadata = metadata;
        this.capacity = metadata.getInput().getCapacity();
    }

    @Override
    public Optional<Component> produce() {
        if (planeCount > 0 && sellStrategy.shouldSell()) {
            sellPlane();
        }

        return Optional.empty();
    }

    @Override
    public void processInput(Component input) {
        planeCount++;
        updateState();
        notifyObservers();
    }

    @Override
    public BuildingMetadata getMetadata() {
        return metadata;
    }

    public void setSellStrategy(SellStrategy sellStrategy) {
        this.sellStrategy = sellStrategy;
    }

    private void updateState() {
        float fullRatio = planeCount / (float) capacity;
        if (fullRatio >= 1) {
            state = BuildingState.FULL;
        } else if (fullRatio >= (2 / 3f)) {
            state = BuildingState.TWO_THIRD;
        } else if (fullRatio >= (1 / 3f)) {
            state = BuildingState.ONE_THIRD;
        } else {
            state = BuildingState.EMPTY;
        }
    }

    private void sellPlane() {
        // Vend un avion
        planeCount--;
        updateState();
        notifyObservers();
    }

    public void attach(WarehouseObserver observer) {
        observers.add(observer);
    }

    public void dettach(WarehouseObserver observer) {
        observers.remove(observer);
    }

    public void notifyObservers() {
        observers.forEach(o -> o.updateWarehouseState(state));
    }

    @Override
    public String toString() {
        boolean isFull = planeCount >= capacity;

        return String.format("[planes: %d, full: %s]", planeCount, isFull);
    }
}
